# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil
# ---------------------------------------------------------------------
# Copyright (c) 2013-2017 Merchise Autrement [~º/~] and Contributors
# Copyright (c) 2012 Medardo Rodríguez
# All rights reserved.
#
# Author: Medardo Rodriguez
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created: 2012-03-23

'''Collection of disparate utilities.

`xoutil` is essentially an extension to the Python's standard library, it does
not turn into a full framework, but it's very useful to be used from a
diversity of scenarios where compatibility is an important issue.

'''

# TODO: Deprecate this import
from .symbols import Unset, Undefined, Ignored, Invalid
